# -*- coding: utf-8 -*-
# vim: set et sts=2:
##############################################################################
#
# Copyright (c) 2010 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import argparse
import gdbm
import json
import logging
import logging.handlers
import os
import Queue
import select
import StringIO
import socket
import signal
import subprocess

cleanup_data = {}

def cleanup(signum=None, frame=None):
  cleanup_functions = dict(
    sockets=lambda sock: sock.close(),
    subprocesses=lambda process: process.terminate(),
    paths=lambda filename: os.unlink(filename),
  )
  for data, function in cleanup_functions.iteritems():
    for item in cleanup_data.get(data, []):
      # XXX will these lists ever have more than 1 element??
      # Swallow everything !
      try:
        function(item)
      except:
        pass

signal.signal(signal.SIGTERM, cleanup)

class TaskRunner(object):

  def __init__(self):
    self._task = None
    self._command = None
    self._time = None

  def has_running_task(self):
    return self._task is not None and self._task.poll() is None

  def had_previous_task(self):
    return self._task is not None and self._task.poll() is not None

  def get_previous_command(self):
    if not self.has_running_task():
      return self._command
    else:
      return None

  def get_previous_timestamp(self):
    if not self.has_running_task():
      return self._time
    else:
      return None

  def get_previous_returncode(self):
    return self._task.poll()

  def flush(self):
    self._task = None
    self._command = None

  def run(self, command, time):
    self._time = time
    self._command = command
    self._task = subprocess.Popen([command], stdin=subprocess.PIPE,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  close_fds=True)
    self._task.stdin.flush()
    self._task.stdin.close()
    cleanup_data.update(subprocesses=[self._task])

  def fd(self):
    if not self.has_running_task():
      raise KeyError("No task is running.")
    return self._task.stdout.fileno()

def main():
  parser = argparse.ArgumentParser(
    description="Run a single threaded execution queue.")
  parser.add_argument('--database', nargs=1, required=True,
                      help="Path to the database where the last "
                           "calls are stored")
  parser.add_argument('--loglevel', nargs=1,
                      default='INFO',
                      choices=[i for i in logging._levelNames
                               if isinstance(i, str)],
                      required=False)
  parser.add_argument('-l', '--logfile', nargs=1, required=True,
                      help="Path to the log file.")
  parser.add_argument('-t', '--timeout', nargs=1, required=False,
                      dest='timeout', type=int, default=3)
  parser.add_argument('socket', help="Path to the unix socket")

  args = parser.parse_args()

  socketpath = args.socket
  level = logging._levelNames.get(args.loglevel[0], logging.INFO)

  logger = logging.getLogger("EQueue")
  # Natively support logrotate
  handler = logging.handlers.WatchedFileHandler(args.logfile[0], mode='a')
  logger.setLevel(level)
  formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
  handler.setFormatter(formatter)
  logger.addHandler(handler)

  unixsocket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

  unixsocket.bind(socketpath)
  logger.debug("Bind on %r", socketpath)
  unixsocket.listen(2)
  logger.debug("Listen on socket")
  unixsocketfd = unixsocket.fileno()

  db = gdbm.open(args.database[0], 'cs', 0700)

  logger.debug("Open timestamp database")

  cleanup_data.update(sockets=[unixsocket])
  cleanup_data.update(paths=[socketpath])

  logger.info("Starting server on %r", socketpath)

  task_queue = Queue.Queue()
  task_running = TaskRunner()

  try:
    rlist = [unixsocketfd]

    while True:
      rlist_s, wlist_s, xlist_s = select.select(rlist, [], [])

      if unixsocketfd in rlist_s:
        conn, addr = unixsocket.accept()
        logger.debug("Connection with file descriptor %d", conn.fileno())

        conn.settimeout(args.timeout)

        request_string = StringIO.StringIO()
        segment = None
        try:
          while segment != '':
            segment = conn.recv(1024)
            request_string.write(segment)
        except socket.timeout:
          pass

        command = '127'
        try:
          request = json.loads(request_string.getvalue())
          timestamp = request['timestamp']
          command = str(request['command'])
          task_queue.put([command, timestamp])
          logger.info("New command %r at %s", command, timestamp)

        except (ValueError, IndexError) :
          logger.warning("Error during the unserialization of json "
                         "message of %r file descriptor. The message "
                         "was %r", conn.fileno(), request_string.getvalue())

        try:
          conn.send(command)
          conn.close()
        except:
          logger.warning("Couldn't respond to %r", conn.fileno())

      rlist = [unixsocketfd]
      if not task_running.has_running_task():
        if task_running.had_previous_task():
          previous_command = task_running.get_previous_command()
          if task_running.get_previous_returncode() != 0:
            logger.warning("%s finished with non zero status.",
                           previous_command)
          else:
            logger.info("%s finished successfully.", previous_command)
          task_running.flush()
          db[previous_command] = str(task_running.get_previous_timestamp())

        try:
          while True:
            command, timestamp = task_queue.get(False)
            if command not in db or timestamp > int(db[command]):
              logger.info("Running %s", command)
              task_running.run(command, timestamp)
              break
            else:
              logger.info("%s already runned.", command)
        except Queue.Empty:
          logger.info("Task queue is empty. Nothing to do...")

      try:
        rlist.append(task_running.fd())
      except KeyError:
        pass

  finally:
    cleanup()

if __name__ == '__main__':
  main()




# -*- coding: utf-8 -*-
import os
import sys
import views
import unittest
import tempfile
from optparse import OptionParser, Option
import ConfigParser
import shutil
from gittools import cloneRepo, gitStatus, switchBranch, addBranch, getDiff, \
     gitPush, gitPull
from flaskext.auth import Auth, AuthUser, login_required, logout
from utils import *
import json
from datetime import timedelta

class Config:
  def setConfig(self):
    """
    Set options given by parameters.
    """
    self.configuration_file_path = os.path.abspath(os.environ.get('CONFIG_FILE_PATH'))

    # Load configuration file
    configuration_parser = ConfigParser.SafeConfigParser()
    configuration_parser.read(self.configuration_file_path)
    # Merges the arguments and configuration

    for section in ("slaprunner", "slapos", "slapproxy", "slapformat",
                    "sshkeys_authority", "gitclient", "cloud9_IDE"):
      configuration_dict = dict(configuration_parser.items(section))
      for key in configuration_dict:
        if not getattr(self, key, None):
          setattr(self, key, configuration_dict[key])

class SlaprunnerTestCase(unittest.TestCase):

  def setUp(self):
    """Initialize slapos webrunner here"""
    views.app.config['TESTING'] = True
    self.users = ["slapuser", "slappwd", "slaprunner@nexedi.com", "SlapOS web runner"]
    self.updateUser = ["newslapuser", "newslappwd", "slaprunner@nexedi.com", "SlapOS web runner"]
    self.rcode = "41bf2657"
    #create slaprunner configuration
    config = Config()
    config.setConfig()
    workdir = os.path.join(config.runner_workdir, 'project')
    views.app.config.update(**config.__dict__)
    #update or create all runner base directory to test_dir
    if not os.path.exists(workdir):
      os.mkdir(workdir)
    views.app.config.update(
      software_log=config.software_root.rstrip('/') + '.log',
      instance_log=config.instance_root.rstrip('/') + '.log',
      workspace = workdir,
      instance_profile='instance.cfg',
      software_profile='software.cfg',
      SECRET_KEY="123456",
      PERMANENT_SESSION_LIFETIME=timedelta(days=31),
    )
    self.app = views.app.test_client()
    self.app.config = views.app.config
    #Create password recover code
    rpwd = open(os.path.join(views.app.config['etc_dir'], '.rcode'), 'w')
    rpwd.write(self.rcode)
    rpwd.close()

  def tearDown(self):
    """Remove all test data"""
    os.unlink(os.path.join(self.app.config['etc_dir'], '.rcode'))
    project = os.path.join(self.app.config['etc_dir'], '.project')
    sr_data = project = os.path.join(self.app.config['etc_dir'], '.softdata')
    proxy = os.path.join(self.app.config['etc_dir'], '.proxy.pid')
    slapgrid_cp = os.path.join(self.app.config['etc_dir'], '.slapgrid-cp.pid')
    slapgrid_sr = os.path.join(self.app.config['etc_dir'], '.slapgrid-sr.pid')
    users = os.path.join(self.app.config['etc_dir'], '.users')
    if os.path.exists(users):
      os.unlink(users)
    if os.path.exists(project):
      os.unlink(project)
    if os.path.exists(sr_data):
      os.unlink(sr_data)
    if os.path.exists(proxy):
      os.unlink(proxy)
    if os.path.exists(slapgrid_cp):
      os.unlink(slapgrid_cp)
    if os.path.exists(slapgrid_sr):
      os.unlink(slapgrid_sr)
    if os.path.exists(self.app.config['workspace']):
      shutil.rmtree(self.app.config['workspace'])
    if os.path.exists(self.app.config['software_root']):
      shutil.rmtree(self.app.config['software_root'])
    if os.path.exists(self.app.config['instance_root']):
      shutil.rmtree(self.app.config['instance_root'])


  #Helpers
  def loadJson(self, response):
    return json.loads(response.data)

  def configAccount(self, username, password, email, name, rcode):
    """Helper for configAccoun"""
    return self.app.post('/configAccount', data=dict(
            username=username,
            password=password,
            email=email,
            name=name,
            rcode=rcode
          ), follow_redirects=True)
  def configAccount(self, username, password, email, name, rcode):
    """Helper for configAccoun"""
    return self.app.post('/configAccount', data=dict(
            username=username,
            password=password,
            email=email,
            name=name,
            rcode=rcode
          ), follow_redirects=True)

  def login(self, username, password):
    """Helper for Login method"""
    return self.app.post('/doLogin', data=dict(
            clogin=username,
            cpwd=password
          ), follow_redirects=True)

  def setAccount(self):
    """Initialize user account and log user in"""
    response = self.loadJson(self.configAccount(self.users[0], self.users[1],
                  self.users[2], self.users[3], self.rcode))
    response2 = self.loadJson(self.login(self.users[0], self.users[1]))
    self.assertEqual(response['result'], "")
    self.assertEqual(response2['result'], "")

  def logout(self):
    """Helper for Logout current user"""
    return self.app.get('/dologout', follow_redirects=True)

  def updateAccount(self, newaccount, rcode):
    """Helper for update user account data"""
    return self.app.post('/updateAccount', data=dict(
            username=newaccount[0],
            password=newaccount[1],
            email=newaccount[2],
            name=newaccount[3],
            rcode=rcode
          ), follow_redirects=True)


  #Begin test case here
  def test_wrong_login(self):
    """Test Login user before create session. This should return error value"""
    response = self.login(self.users[0], self.users[1])
    #redirect to config account page
    assert "<h2 class='title'>Your personal informations</h2><br/>" in response.data

  def test_configAccount(self):
    """For the first lauch of slaprunner user need do create first account"""
    result = self.configAccount(self.users[0], self.users[1], self.users[2],
                  self.users[3], self.rcode)
    response = self.loadJson(result)
    self.assertEqual(response['code'], 1)
    account = getSession(self.app.config)
    self.assertEqual(account, self.users)

  def test_login_logout(self):
    """test login with good and wrong values, test logout"""
    response = self.loadJson(self.configAccount(self.users[0], self.users[1],
                  self.users[2], self.users[3], self.rcode))
    self.assertEqual(response['result'], "")
    result = self.loadJson(self.login(self.users[0], "wrongpwd"))
    self.assertEqual(result['result'], "Login or password is incorrect, please check it!")
    resultwr = self.loadJson(self.login("wronglogin", "wrongpwd"))
    self.assertEqual(resultwr['result'], "Login or password is incorrect, please check it!")
    #try now with true values
    resultlg = self.loadJson(self.login(self.users[0], self.users[1]))
    self.assertEqual(resultlg['result'], "")
    #after login test logout
    result = self.logout()
    assert "<h2>Login to Slapos Web Runner</h2>" in result.data

  def test_updateAccount(self):
    """test Update accound, this need user to loging in"""
    self.setAccount()
    response = self.loadJson(self.updateAccount(self.updateUser, self.rcode))
    self.assertEqual(response['code'], 1)
    result = self.logout()
    assert "<h2>Login to Slapos Web Runner</h2>" in result.data
    #retry login with new values
    response = self.loadJson(self.login(self.updateUser[0], self.updateUser[1]))
    self.assertEqual(response['result'], "")
    #log out now!
    self.logout()


def main():
  argv = ['', 'test_wrong_login', 'test_configAccount',
          'test_login_logout', 'test_updateAccount']
  unittest.main(module=SlaprunnerTestCase, argv=argv)

if __name__ == '__main__':
  main()


# -*- coding: utf-8 -*-

import slapos.slap
import time
import subprocess
import os
from xml_marshaller import xml_marshaller
from xml.dom import minidom
import re
import urllib
from flask import jsonify
import shutil
import string
import hashlib
import signal
import multiprocessing

# Setup default flask (werkzeug) parser
import logging
logger = logging.getLogger('werkzeug')


class Popen(subprocess.Popen):
  def __init__(self, *args, **kwargs):
    kwargs['stdin'] = subprocess.PIPE
    kwargs['stderr'] = subprocess.STDOUT
    kwargs.setdefault('stdout', subprocess.PIPE)
    kwargs.setdefault('close_fds', True)
    subprocess.Popen.__init__(self, *args, **kwargs)
    self.stdin.flush()
    self.stdin.close()
    self.stdin = None

html_escape_table = {
  "&": "&amp;",
  '"': "&quot;",
  "'": "&apos;",
  ">": "&gt;",
  "<": "&lt;",
}

def html_escape(text):
  """Produce entities within text."""
  return "".join(html_escape_table.get(c,c) for c in text)

def getSession(config):
  """
  Get the session data of current user.
  Returns:
    a list of user informations or False if fail to read data.
  """
  user_path = os.path.join(config['etc_dir'], '.users')
  user = ""
  if os.path.exists(user_path):
    f = open(user_path, 'r')
    user = f.read().split(';')
    f.close()
  if type(user) == type(""):
    return False
  return user

def saveSession(config, account):
  """
  Save account information for the current user

  Args:
    config: Slaprunner configuration
    session: Flask session
    account: New session data to be save

  Returns:
    True if all goes well or str (error message) if fail
  """
  user = os.path.join(config['etc_dir'], '.users')
  backup = False
  try:
    if os.path.exists(user):
      f = open(user, 'r')
      #backup previous data
      data = f.read()
      open(user+'.back', 'w').write(data)
      f.close()
      backup = True
      if not account[1]:
        account[1] = data.split(';')[1]
    #save new account data
    open(user, 'w').write((';'.join(account)).encode("utf-8"))
    return True
  except Exception, e:
    try:
      if backup:
        os.remove(user)
        os.rename(user+'.back', user)
    except:
      pass
    return str(e)

def getCurrentSoftwareReleaseProfile(config):
  """
  Returns used Software Release profile as a string.
  """
  try:
    software_folder = open(
        os.path.join(config['etc_dir'], ".project")).read()
    return realpath(
        config, os.path.join(software_folder, config['software_profile']))
  except:
    return False

def requestInstance(config, software_type=None):
  """
  Request the main instance of our environment
  """
  if not software_type:
    software_type = None

  slap = slapos.slap.slap()
  profile = getCurrentSoftwareReleaseProfile(config)
  slap.initializeConnection(config['master_url'])

  param_path = os.path.join(config['etc_dir'], ".parameter.xml")
  xml_result = readParameters(param_path)
  partition_parameter_kw = None
  if type(xml_result) != type('') and xml_result.has_key('instance'):
    partition_parameter_kw = xml_result['instance']

  return slap.registerOpenOrder().request(
      profile,
      partition_reference=getSoftwareReleaseName(config),
      partition_parameter_kw=partition_parameter_kw,
      software_type=software_type,
      filter_kw=None,
      state=None,
      shared=False)

def updateProxy(config):
  """
  Configure Slapos Node computer and partitions.
  Send current Software Release to Slapproxy for compilation and deployment.
  """
  if not os.path.exists(config['instance_root']):
    os.mkdir(config['instance_root'])
  slap = slapos.slap.slap()
  profile = getCurrentSoftwareReleaseProfile(config)

  slap.initializeConnection(config['master_url'])
  slap.registerSupply().supply(profile, computer_guid=config['computer_id'])
  computer = slap.registerComputer(config['computer_id'])
  prefix = 'slappart'
  slap_config = {
 'address': config['ipv4_address'],
 'instance_root': config['instance_root'],
 'netmask': '255.255.255.255',
 'partition_list': [],
 'reference': config['computer_id'],
 'software_root': config['software_root']}
  for i in xrange(0, int(config['partition_amount'])):
    partition_reference = '%s%s' % (prefix, i)
    partition_path = os.path.join(config['instance_root'], partition_reference)
    if not os.path.exists(partition_path):
      os.mkdir(partition_path)
    os.chmod(partition_path, 0750)
    slap_config['partition_list'].append({'address_list': [{'addr': config['ipv4_address'],
                                       'netmask': '255.255.255.255'},
                                      {'addr': config['ipv6_address'],
                                       'netmask': 'ffff:ffff:ffff::'},
                      ],
                     'path': partition_path,
                     'reference': partition_reference,
                     'tap': {'name': partition_reference},
                     })
  computer.updateConfiguration(xml_marshaller.dumps(slap_config))
  return True

def readPid(file):
  """Read process pid from file `file`"""
  if os.path.exists(file):
    data = open(file).read().strip()
    try:
      return int(data)
    except Exception:
      return 0
  return 0


def writePid(file, pid):
  """Save process pid into a file `file`"""
  open(file, 'w').write(str(pid))

def updateInstanceParameter(config, software_type=None):
  """
  Reconfigure Slapproxy to re-deploy current Software Instance with parameters.

  Args:
    config: Slaprunner configuration.
    software_type: reconfigure Software Instance with software type.
  """
  if not (updateProxy(config) and requestInstance(config, software_type)):
    return False

def startProxy(config):
  """Start Slapproxy server"""
  proxy_pid = os.path.join(config['etc_dir'], 'proxy.pid')
  pid = readPid(proxy_pid)
  running = False
  if pid:
    try:
      os.kill(pid, 0)
    except Exception:
      pass
    else:
      running = True
  if not running:
    proxy = Popen([config['slapproxy'], config['configuration_file_path']])
    proxy_pid = os.path.join(config['etc_dir'], 'proxy.pid')
    writePid(proxy_pid, proxy.pid)
    time.sleep(5)


def stopProxy(config):
  """Stop Slapproxy server"""
  pid = readPid(os.path.join(config['etc_dir'], 'proxy.pid'))
  if pid:
    try:
      os.kill(pid)
    except:
      pass


def removeProxyDb(config):
  """Remove Slapproxy database, this is use to initialize proxy for example when
    configuring new Software Release"""
  if os.path.exists(config['database_uri']):
    os.unlink(config['database_uri'])

def isSoftwareRunning(config):
  """
    Return True if slapgrid-sr is still running and false if slapgrid if not
  """
  slapgrid_pid = os.path.join(config['etc_dir'], 'slapgrid-sr.pid')
  pid = readPid(slapgrid_pid)
  if pid:
    try:
      os.kill(pid, 0)
    except Exception:
      running = False
    else:
      running = True
  else:
    running = False
  return running


def runSoftwareWithLock(config):
  """
    Use Slapgrid to compile current Software Release and wait until
    compilation is done
  """
  slapgrid_pid = os.path.join(config['etc_dir'], 'slapgrid-sr.pid')
  if not isSoftwareRunning(config):
    if not os.path.exists(config['software_root']):
      os.mkdir(config['software_root'])
    stopProxy(config)
    removeProxyDb(config)
    startProxy(config)
    logfile = open(config['software_log'], 'w')
    if not updateProxy(config):
      return False
    # Accelerate compilation by setting make -jX
    environment = os.environ.copy()
    environment['MAKEFLAGS'] = '-j%r' % multiprocessing.cpu_count()
    slapgrid = Popen([config['slapgrid_sr'], '-vc',
        config['configuration_file_path'], '--now', '--develop'],
        stdout=logfile, env=environment)
    writePid(slapgrid_pid, slapgrid.pid)
    slapgrid.wait()
    #Saves the current compile software for re-use
    config_SR_folder(config)
    return True
  return False

def config_SR_folder(config):
  """Create a symbolik link for each folder in software folder. That allow
    user to customize software release folder"""
  list = []
  config_name = 'slaprunner.config'
  for path in os.listdir(config['software_link']):
    cfg_path = os.path.join(config['software_link'], path, config_name)
    if os.path.exists(cfg_path):
      cfg = open(cfg_path, 'r').read().split("#")
      if len(cfg) != 2:
        continue #there is a broken config file
      list.append(cfg[1])
  folder_list = os.listdir(config['software_root'])
  if len(folder_list) < 1:
    return
  curent_project = open(os.path.join(config['etc_dir'], ".project"),
                        'r').read()
  projects = curent_project.split("/")
  name = projects[len(projects) - 2]
  for folder in folder_list:
    if folder in list:
      continue #this folder is already registered
    else:
      if not os.path.exists(os.path.join(config['software_link'], name)):
        destination = os.path.join(config['software_link'], name)
      else:
        destination = os.path.join(config['software_link'], folder)
      source = os.path.join(config['software_root'], folder)
      cfg = os.path.join(destination, config_name)
      #create symlink
      os.symlink(source, destination)
      #write config file
      cf = open(cfg, 'w')
      cf.write(curent_project+"#"+folder)
      cf.close()

def loadSoftwareRList(config):
  """Return list (of dict) of Software Release from symbolik SR folder"""
  list = []
  config_name = 'slaprunner.config'
  for path in os.listdir(config['software_link']):
    cfg_path = os.path.join(config['software_link'], path, config_name)
    if os.path.exists(cfg_path):
      cfg = open(cfg_path, 'r').read().split("#")
      if len(cfg) != 2:
        continue #there is a broken config file
      list.append(dict(md5=cfg[1], path=cfg[0], title=path))
  return list

def isInstanceRunning(config):
  """
    Return True if slapgrid-cp is still running and false if slapgrid if not
  """
  slapgrid_pid = os.path.join(config['etc_dir'], 'slapgrid-cp.pid')
  pid = readPid(slapgrid_pid)
  if pid:
    try:
      os.kill(pid, 0)
    except Exception:
      running = False
    else:
      running = True
  else:
    running = False
  return running

def killRunningSlapgrid(config, ptype):
  """Kill slapgrid process and all running children process"""
  slapgrid_pid = os.path.join(config['etc_dir'], ptype)
  pid = readPid(slapgrid_pid)
  if pid:
      recursifKill([pid])
  else:
    return False

def recursifKill(pids):
  """Try to kill a list of proccess by the given pid list"""
  if pids == []:
    return
  else:
    for pid in pids:
      ppids = pidppid(pid)
      try:
        os.kill(pid, signal.SIGKILL) #kill current process
      except Exception:
              pass
      recursifKill(ppids) #kill all children of this process

def pidppid(pid):
  """get the list of the children pids of a process `pid`"""
  proc = Popen('ps -o pid,ppid ax | grep "%d"' % pid, shell=True,
               stdout=subprocess.PIPE)
  ppid  = [x.split() for x in proc.communicate()[0].split("\n") if x]
  return list(int(p) for p, pp in ppid if int(pp) == pid)

def runInstanceWithLock(config):
  """
    Use Slapgrid to deploy current Software Release and wait until
    deployment is done.
  """
  slapgrid_pid = os.path.join(config['etc_dir'], 'slapgrid-cp.pid')
  if not isInstanceRunning(config):
    startProxy(config)
    logfile = open(config['instance_log'], 'w')
    if not (updateProxy(config) and requestInstance(config)):
      return False
    svcStopAll(config) #prevent lost control of process
    slapgrid = Popen([config['slapgrid_cp'], '-vc',
        config['configuration_file_path'], '--now'],
        stdout=logfile)
    writePid(slapgrid_pid, slapgrid.pid)
    slapgrid.wait()
    return True
  return False

def getProfilePath(projectDir, profile):
  """
  Return the path of the current Software Release `profile`

  Args:
    projectDir: Slaprunner workspace location.
    profile: file to search into the workspace.

  Returns:
    String, path of current Software Release profile
  """
  if not os.path.exists(os.path.join(projectDir, ".project")):
    return False
  projectFolder = open(os.path.join(projectDir, ".project")).read()
  return os.path.join(projectFolder, profile)

def getSlapStatus(config):
  """Return all Slapos Partitions with associate informations"""
  slap = slapos.slap.slap()
  slap.initializeConnection(config['master_url'])
  partition_list = []
  computer = slap.registerComputer(config['computer_id'])
  try:
    for partition in computer.getComputerPartitionList():
      # Note: Internal use of API, as there is no reflexion interface in SLAP
      partition_list.append((partition.getId(), partition._connection_dict.copy()))
  except Exception:
    pass
  if partition_list:
    for i in xrange(0, int(config['partition_amount'])):
      slappart_id = '%s%s' % ("slappart", i)
      if not [x[0] for x in partition_list if slappart_id==x[0]]:
        partition_list.append((slappart_id, []))
  return partition_list

def svcStopAll(config):
  """Stop all Instance process on this computer"""
  return Popen([config['supervisor'], config['configuration_file_path'],
                'shutdown']).communicate()[0]

def removeInstanceRoot(config):
  """Clean instance directory and stop all its running process"""
  if os.path.exists(config['instance_root']):
    svcStopAll(config)
    for root, dirs, files in os.walk(config['instance_root']):
      for fname in dirs:
        fullPath = os.path.join(root, fname)
        if not os.access(fullPath, os.W_OK) :
          # Some directories may be read-only, preventing to remove files in it
          os.chmod(fullPath, 0744)
    shutil.rmtree(config['instance_root'])

def getSvcStatus(config):
  """Return all Softwares Instances process Informations"""
  result = Popen([config['supervisor'], config['configuration_file_path'],
                  'status']).communicate()[0]
  regex = "(^unix:.+\.socket)|(^error:).*$"
  supervisord = []
  for item in result.split('\n'):
    if item.strip() != "":
      if re.search(regex, item, re.IGNORECASE) == None:
        supervisord.append(re.split('[\s,]+', item))
      else:
        return [] #ignore because it is an error message
  return supervisord

def getSvcTailProcess(config, process):
  """Get log for the specifie process

  Args:
    config: Slaprunner configuration
    process: process name. this value is pass to supervisord.
  Returns:
    a string that contains the log of the process.
  """
  return Popen([config['supervisor'], config['configuration_file_path'],
                "tail", process]).communicate()[0]

def svcStartStopProcess(config, process, action):
  """Send start or stop process command to supervisord

  Args:
    config: Slaprunner configuration.
    process: process to start or stop.
    action: current state which is used to generate the new process state.
  """
  cmd = {"RESTART":"restart", "STOPPED":"start", "RUNNING":"stop", "EXITED":"start", "STOP":"stop"}
  return Popen([config['supervisor'], config['configuration_file_path'],
                cmd[action], process]).communicate()[0]

def getFolderContent(config, folder):
  """
  Read all file and folder into specified directory

  Args:
    config: Slaprunner configuration.
    folder: the directory to read.

  Returns:
    Html formated string or error message when fail.
  """
  r=['<ul class="jqueryFileTree" style="display: none;">']
  try:
    folder = str(folder)
    r=['<ul class="jqueryFileTree" style="display: none;">']
    d=urllib.unquote(folder)
    realdir = realpath(config, d)
    if not realdir:
      r.append('Could not load directory: Permission denied')
      ldir = []
    else:
      ldir = sorted(os.listdir(realdir), key=str.lower)
    for f in ldir:
      if f.startswith('.'): #do not displays this file/folder
        continue
      ff=os.path.join(d,f)
      if os.path.isdir(os.path.join(realdir,f)):
        r.append('<li class="directory collapsed"><a href="#%s" rel="%s/">%s</a></li>' % (ff, ff,f))
      else:
        e=os.path.splitext(f)[1][1:] # get .ext and remove dot
        r.append('<li class="file ext_%s"><a href="#%s" rel="%s">%s</a></li>' % (e, ff,ff,f))
    r.append('</ul>')
  except Exception,e:
    r.append('Could not load directory: %s' % str(e))
  r.append('</ul>')
  return jsonify(result=''.join(r))

def getFolder(config, folder):
  """
  Read list of folder for the specified directory

  Args:
    config: Slaprunner configuration.
    folder: the directory to read.

  Returns:
    Html formated string or error message when fail.
  """
  r=['<ul class="jqueryFileTree" style="display: none;">']
  try:
    folder = str(folder)
    r=['<ul class="jqueryFileTree" style="display: none;">']
    d=urllib.unquote(folder)
    realdir = realpath(config, d)
    if not realdir:
      r.append('Could not load directory: Permission denied')
      ldir = []
    else:
      ldir = sorted(os.listdir(realdir), key=str.lower)
    for f in ldir:
      if f.startswith('.'): #do not display this file/folder
        continue
      ff=os.path.join(d,f)
      if os.path.isdir(os.path.join(realdir,f)):
        r.append('<li class="directory collapsed"><a href="#%s" rel="%s/">%s</a></li>' % (ff, ff, f))
    r.append('</ul>')
  except Exception,e:
    r.append('Could not load directory: %s' % str(e))
  r.append('</ul>')
  return jsonify(result=''.join(r))

def getProjectList(folder):
  """Return the list of projet (folder) into the workspace

  Agrs:
    folder: path of the workspace
  Returns:
    a list that contains each folder name.
  """
  project = []
  project_list = sorted(os.listdir(folder), key=str.lower)
  for elt in project_list:
    if os.path.isdir(os.path.join(folder, elt)):
      project.append(elt)
  return project

def configNewSR(config, projectpath):
  """Configure a Software Release as current Software Release

  Args:
    config: slaprunner configuration
    projectpath: path of the directory that contains the software realease to configure
  Returns:
    True if all is done well, otherwise return false.
  """
  folder = realpath(config, projectpath)
  if folder:
    if isInstanceRunning(config):
      killRunningSlapgrid(config, "slapgrid-cp.pid")
    if isSoftwareRunning(config):
      killRunningSlapgrid(config, "slapgrid-sr.pid")
    stopProxy(config)
    removeProxyDb(config)
    startProxy(config)
    removeInstanceRoot(config)
    param_path = os.path.join(config['etc_dir'], ".parameter.xml")
    if os.path.exists(param_path):
      os.remove(param_path)
    open(os.path.join(config['etc_dir'], ".project"), 'w').write(projectpath)
    return True
  else:
    return False

def newSoftware(folder, config, session):
  """
  Create a new Software Release folder with default profiles

  Args:
    folder: directory of the new software release
    config: slraprunner configuration
    session: Flask session directory"""
  json = ""
  code = 0
  basedir = config['etc_dir']
  try:
    folderPath = realpath(config, folder, check_exist=False)
    if folderPath and not os.path.exists(folderPath):
      os.mkdir(folderPath)
      #load software.cfg and instance.cfg from http://git.erp5.org
      software = "http://git.erp5.org/gitweb/slapos.git/blob_plain/HEAD:/software/lamp-template/software.cfg"
      softwareContent = ""
      try:
        softwareContent = urllib.urlopen(software).read()
      except:
        #Software.cfg and instance.cfg content will be empty
        pass
      open(os.path.join(folderPath, config['software_profile']), 'w').write(softwareContent)
      open(os.path.join(folderPath, config['instance_profile']), 'w').write("")
      open(os.path.join(basedir, ".project"), 'w').write(folder + "/")
      session['title'] = getProjectTitle(config)
      code = 1
    else:
      json = "Bad folder or Directory '" + folder + \
        "' already exist, please enter a new name for your software"
  except Exception, e:
    json = "Can not create your software, please try again! : " + str(e)
    if os.path.exists(folderPath):
      shutil.rmtree(folderPath)
  return jsonify(code=code, result=json)

def checkSoftwareFolder(path, config):
  """Check id `path` is a valid Software Release folder"""
  realdir = realpath(config, path)
  if realdir and os.path.exists(os.path.join(realdir, config['software_profile'])):
    return jsonify(result=path)
  return jsonify(result="")

def getProjectTitle(config):
  """Generate the name of the current software Release (for slaprunner UI)"""
  conf = os.path.join(config['etc_dir'], ".project")
  if os.path.exists(conf):
    project = open(conf, "r").read().split("/")
    software = project[len(project) - 2]
    return software + " (" + string.join(project[:(len(project) - 2)], '/') + ")"
  return "No Profile"

def getSoftwareReleaseName(config):
  """Get the name of the current Software Release"""
  sr_profile = os.path.join(config['etc_dir'], ".project")
  if os.path.exists(sr_profile):
    project = open(sr_profile, "r").read().split("/")
    software = project[len(project) - 2]
    return software.replace(' ', '_')
  return "No_name"

def removeSoftwareByName(config, md5, folderName):
  """Remove all content of the software release specified by md5

  Args:
    config: slaprunner configuration
    foldername: the link name given to the software release
    md5: the md5 filename given by slapgrid to SR folder"""
  if isSoftwareRunning(config) or isInstanceRunning(config):
    raise Exception("Software installation or instantiation in progress, cannot remove")
  path = os.path.join(config['software_root'], md5)
  linkpath = os.path.join(config['software_link'], folderName)
  if not os.path.exists(path):
    raise Exception("Cannot remove software Release: No such file or directory")
  if not os.path.exists(linkpath):
    raise Exception("Cannot remove software Release: No such file or directory %s" %
                    ('software_root/'+folderName))
  svcStopAll(config)
  os.unlink(linkpath)
  shutil.rmtree(path)
  return loadSoftwareRList(config)

def tail(f, lines=20):
  """
  Returns the last `lines` lines of file `f`. It is an implementation of tail -f n.
  """
  BUFSIZ = 1024
  f.seek(0, 2)
  bytes = f.tell()
  size = lines + 1
  block = -1
  data = []
  while size > 0 and bytes > 0:
      if bytes - BUFSIZ > 0:
          # Seek back one whole BUFSIZ
          f.seek(block * BUFSIZ, 2)
          # read BUFFER
          data.insert(0, f.read(BUFSIZ))
      else:
          # file too small, start from begining
          f.seek(0,0)
          # only read what was not read
          data.insert(0, f.read(bytes))
      linesFound = data[0].count('\n')
      size -= linesFound
      bytes -= BUFSIZ
      block -= 1
  return string.join(''.join(data).splitlines()[-lines:], '\n')

def readFileFrom(f, lastPosition):
  """
  Returns the last lines of file `f`, from position lastPosition.
  and the last position
  """
  BUFSIZ = 1024
  f.seek(0, 2)
  bytes = f.tell()
  block = -1
  data = ""
  length = bytes
  if lastPosition <= 0 and length > 30000:
    lastPosition = length-30000
  size = bytes - lastPosition
  while bytes > lastPosition:
    if abs(block*BUFSIZ) <= size:
      # Seek back one whole BUFSIZ
      f.seek(block * BUFSIZ, 2)
      data = f.read(BUFSIZ) + data
    else:
      margin = abs(block*BUFSIZ) - size
      if length < BUFSIZ:
        f.seek(0,0)
      else:
        seek = block * BUFSIZ + margin
        f.seek(seek, 2)
      data = f.read(BUFSIZ - margin) + data
    bytes -= BUFSIZ
    block -= 1
  f.close()
  return {"content":data, "position":length}

def isText(file):
  """Return True if the mimetype of file is Text"""
  if not os.path.exists(file):
    return False
  text_range = ''.join(map(chr, [7,8,9,10,12,13,27] + range(0x20, 0x100)))
  is_binary_string = lambda bytes: bool(bytes.translate(None, text_range))
  try:
    return not is_binary_string(open(file).read(1024))
  except:
    return False

def md5sum(file):
  """Compute md5sum of `file` and return hexdigest value"""
  if os.path.isdir(file):
    return False
  try:
    fh = open(file, 'rb')
    m = hashlib.md5()
    while True:
      data = fh.read(8192)
      if not data:
        break
      m.update(data)
    return m.hexdigest()
  except:
    return False

def realpath(config, path, check_exist=True):
  """
  Get realpath of path or return False if user is not allowed to access to
  this file.
  """
  split_path = path.split('/')
  key = split_path[0]
  allow_list = {'software_root':config['software_root'], 'instance_root':
                config['instance_root'], 'workspace': config['workspace'],
                'software_link':config['software_link']}
  if allow_list.has_key(key):
    del split_path[0]
    path = os.path.join(allow_list[key], *split_path)
    if check_exist:
      if os.path.exists(path):
        return path
      else:
        return False
    else:
      return path
  return False

def readParameters(path):
  """Read Instance parameters stored into a local file.

  Agrs:
    path: path of the xml file that contains parameters

  Return:
    a dictionnary of instance parameters."""
  if os.path.exists(path):
    try:
      xmldoc = minidom.parse(path)
      object = {}
      for elt in xmldoc.childNodes:
        sub_object = {}
        for subnode in elt.childNodes:
          if subnode.nodeType != subnode.TEXT_NODE:
            sub_object[str(subnode.getAttribute('id'))] = subnode.childNodes[0].data #.decode('utf-8').decode('utf-8')
            object[str(elt.tagName)] = sub_object
      return object
    except Exception, e:
      return str(e)
  else:
    return "No such file or directory: " + path



import ConfigParser
import argparse
import datetime
import httplib
import json
import logging
import os
import random
import sys
import tempfile
import traceback
import time
import xmlrpclib

import slapos.slap
from slapos.grid.utils import setRunning, setFinished

from erp5.util.taskdistribution import TaskDistributionTool, RPCRetry
from erp5.util.taskdistribution import SAFE_RPC_EXCEPTION_LIST


class AutoSTemp(object):
    """
    Create a self-destructing temporary file.
    Uses mkstemp.
    """
    __unlink = os.unlink

    def __init__(self, value):
        fd, self.__name = tempfile.mkstemp()
        os.write(fd, value)
        os.close(fd)

    @property
    def name(self):
        return self.__name

    def __del__(self):
        self.__unlink(self.__name)

SOFTWARE_STATE_UNKNOWN = -1
SOFTWARE_STATE_INSTALLING = 0
SOFTWARE_STATE_INSTALLED = 1
SOFTWARE_STATE_DESTROYING = 2

INSTANCE_STATE_UNKNOWN = -1
INSTANCE_STATE_STARTING = 0
INSTANCE_STATE_STARTED = 1
INSTANCE_STATE_STOPPING = 2
INSTANCE_STATE_STOPPED = 3
INSTANCE_STATE_DESTROYING = 4

TESTER_STATE_INITIAL = -1
TESTER_STATE_NOTHING = 0
TESTER_STATE_SOFTWARE_INSTALLED = 1
TESTER_STATE_INSTANCE_INSTALLED = 2
TESTER_STATE_INSTANCE_STARTED = 4
TESTER_STATE_INSTANCE_UNINSTALLED = 5

class x509Transport(xmlrpclib.Transport):
    """
    Similar to xmlrpclib.SecureTransport, but with actually usable x509
    support.
    """
    def __init__(self, x509, *args, **kw):
        xmlrpclib.Transport.__init__(self, *args, **kw)
        self.__x509 = x509

    def make_connection(self, host):
        if not self._connection or host != self._connection[0]:
            try:
                HTTPSConnection = httplib.HTTPSConnection
            except AttributeError:
                raise NotImplementedError("your version of httplib doesn't "
                    "support HTTPS")
            else:
                chost, self._extra_headers, x509 = self.get_host_info((host,
                    self.__x509))
                self._connection = (host, HTTPSConnection(chost, None, **x509))
        return self._connection[1]

class TestTimeout(Exception):
    pass

# Simple decorator to prevent raise due small
# network failures.
def retryOnNetworkFailure(func):
  def wrapper(*args, **kwargs):
    retry_time = 64
    while True:
      try:
        return func(*args, **kwargs)
      except SAFE_RPC_EXCEPTION_LIST, e:
        print 'Network failure: %s , %s' % (sys.exc_info(), e)

      print 'Retry method %s in %i seconds' % (func, retry_time)
      time.sleep(retry_time)
      retry_time += retry_time >> 1

  wrapper.__name__ = func.__name__
  wrapper.__doc__ = func.__doc__
  return wrapper

class SoftwareReleaseTester(RPCRetry):
    deadline = None
    latest_state = None

    def __init__(self,
                name,
                logger,
                master,
                slap_supply, # slapos supply to manage software release
                slap_order, # slapos open order to manage instance
                url, # software release url
                computer_guid, # computer to use for this test run
                max_install_duration,
                max_uninstall_duration,
                request_kw=None, # instance parameters, if instantiation
                    # testing is desired
                max_request_duration=None,
                max_destroy_duration=None,
            ):
        super(SoftwareReleaseTester, self).__init__(master, 16, logger)
        self.name = name
        self.slap_supply = slap_supply
        self.slap_order = slap_order
        self.url = url
        self.computer_guid = computer_guid
        self.request_kw = request_kw
        self.state = TESTER_STATE_INITIAL
        self.transition_dict = {
            # step function
            # delay
            # next_state
            # software_state
            # instance_state
            TESTER_STATE_INITIAL: (
                lambda t: None,
                None,
                TESTER_STATE_NOTHING,
                None,
                None,
            ),
            TESTER_STATE_NOTHING: (
                lambda t: t.install(),
                max_install_duration,
                request_kw is None and TESTER_STATE_INSTANCE_UNINSTALLED or \
                    TESTER_STATE_SOFTWARE_INSTALLED,
                SOFTWARE_STATE_INSTALLED,
                None,
            ),
            TESTER_STATE_SOFTWARE_INSTALLED: (
                lambda t: t.request(),
                max_request_duration,
                TESTER_STATE_INSTANCE_STARTED,
                None,
                INSTANCE_STATE_STARTED,
            ),
            TESTER_STATE_INSTANCE_STARTED: (
                lambda t: t.destroy(),
                max_destroy_duration,
                TESTER_STATE_INSTANCE_UNINSTALLED,
                None,
                INSTANCE_STATE_UNKNOWN,
            ),
            TESTER_STATE_INSTANCE_UNINSTALLED: (
                lambda t: t.uninstall(),
                max_uninstall_duration,
                None,
                None,
                INSTANCE_STATE_UNKNOWN,
            ),
        }

    def __repr__(self):
        deadline = self.deadline
        if deadline is not None:
            deadline -= time.time()
            deadline = '+%is' % (deadline, )
        return '<%s(state=%s, deadline=%s) at %x>' % (
            self.__class__.__name__, self.state, deadline, id(self))

    @retryOnNetworkFailure
    def _supply(self, state):
        self._logger.info('Supply %s@%s: %s', self.url, self.computer_guid,
            state)
        return self.slap_supply.supply(self.url, self.computer_guid, state)

    @retryOnNetworkFailure
    def _request(self, state):
        self._logger.info('Request %s@%s: %s', self.url, self.name, state)
        self.latest_state = state
        return self.slap_order.request(
            software_release=self.url,
            partition_reference=self.name,
            state=state,
            **self.request_kw
        )

    def _getSoftwareState(self):
        return SOFTWARE_STATE_INSTALLED

    def _getInstanceState(self):
        latest_state = self.latest_state
        self._logger.debug('latest_state = %r', latest_state)

        if latest_state is None:
            return INSTANCE_STATE_UNKNOWN
        try:
            requested = self._request(latest_state)
            if requested._computer_id is None:
                return INSTANCE_STATE_UNKNOWN
        except slapos.slap.ServerError:
            self._logger.exception('Got an error requesting partition for '
                'its state')
            return INSTANCE_STATE_UNKNOWN

        try:
            instance_state = requested.getStatus()
            # the following does NOT take TZ into account
            created_at = datetime.datetime.strptime(instance_state['created_at'], '%a, %d %b %Y %H:%M:%S %Z')
            gmt_now = datetime.datetime(*time.gmtime()[:6])

            self._logger.debug('Instance state: ', instance_state)
            self._logger.debug('Created at: %s (%d)' % (instance_state['created_at'], (gmt_now - created_at).seconds))

            if instance_state['text'].startswith('#error no data found'):
                return INSTANCE_STATE_UNKNOWN
            elif instance_state['text'].startswith('#access') \
                    and (gmt_now - created_at).seconds < 300:
                return INSTANCE_STATE_STARTED
        except slapos.slap.ResourceNotReady:
            return INSTANCE_STATE_UNKNOWN

        return INSTANCE_STATE_UNKNOWN

    def install(self):
        """
        Make software available on computer.
        """
        self._supply('available')

    def uninstall(self):
        """
        Make software unavailable on computer.
        """
        self._supply('destroyed')

    def start(self):
        """
        Request started instance (or starting existing one)
        """
        self._request('started')
    request = start

    def stop(self):
        """
        Request stopped instance (or stopping existing one).
        """
        self._request('stopped')

    def destroy(self):
        """
        Destroy existing instance.
        """
        self._request('destroyed')

    def teardown(self):
        """
        Interrupt a running test sequence, putting it in idle state.
        """
        self._logger.info('Invoking TearDown for %s@%s' % (self.url, self.name))
        if self.request_kw is not None:
            self.destroy()
        self.uninstall()
        self.state = TESTER_STATE_INSTANCE_UNINSTALLED

    def tic(self, now):
        """
        Check for missed deadlines (-> test failure), conditions for moving to
        next state, and actually moving to next state (executing its payload).
        """
        self._logger.debug('TIC')
        deadline = self.deadline
        if deadline < now and deadline is not None:
            raise TestTimeout(self.state)
        _, _, next_state, software_state, instance_state = self.transition_dict[
            self.state]

        if (software_state is None or
                software_state == self._getSoftwareState()) and (
                instance_state is None or
                instance_state == self._getInstanceState()):
            self._logger.debug('Going to state %s (%r, %r)', next_state,
                software_state, instance_state)
            if next_state is None:
                return None
            self.state = next_state
            stepfunc, delay, _, _, _ = self.transition_dict[next_state]
            self.deadline = now + delay
            stepfunc(self)
        return self.deadline

class TestMap(object):
   def __init__(self, test_dict):
       self.test_map_dict = {}
       for key in test_dict:
           target_computer = test_dict[key]["target_computer"]
           if target_computer not in self.test_map_dict:
               self.test_map_dict[target_computer] = [key]
           else:
               self.test_map_dict[target_computer].append(key)

   def getExcludeList(self, computer_id):
       exclude_list = []
       for key in self.test_map_dict:
           if key != computer_id:
               exclude_list.extend(self.test_map_dict[key])
       return set(exclude_list)

   def getComputerList(self):
       return self.test_map_dict.keys()

   def dropComputer(self, computer_id):
       del self.test_map_dict[computer_id]

   def cleanUp(self):
       for key in self.test_map_dict.copy():
          if len(self.test_map_dict[key]) == 0:
              del self.test_map_dict[key]

   def getNextComputer(self, used_computer_list):
       for computer in self.getComputerList():
           if computer not in used_computer_list:
               return computer

       return None

def main():
    """
    Note: This code does not test as much as it monitors.
    The goal is to regularily try to build & instantiate a software release
    on several machines, to monitor vifib stability and SR stability as time
    passes (and things once available online become unavailable).
    Part of this function could be reused to make an actual test bot, testing
    only when actual changes are committed to a software release, to look for
    regressions.

    Note: This code does not connect to any instantiated service, it relies on
    the presence of a promise section to make instantiation fail until promise
    is happy.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--pidfile', '-p', help='pidfile preventing parallel '
        'execution.')
    parser.add_argument('--log', '-l', help='Log file path.')
    parser.add_argument('--verbose', '-v', help='Be verbose.',
        action='store_true')
    parser.add_argument('configuration_file', type=argparse.FileType(),
        help='Slap Test Agent configuration file.')
    # Just to keep strong references to AutoSTemp instances
    key_file_dict = {}
    def asFilenamePair(key, cert):
        # Note: python's ssl support only supports fetching key & cert data
        # from on-disk files. This is why we need to "convert" direct data
        # into file paths, using temporary files.
        cert = cert.strip()
        try:
            temp_key, temp_cert = key_file_dict[cert]
        except KeyError:
            temp_key = AutoSTemp(key.strip())
            temp_cert = AutoSTemp(cert)
            key_file_dict[cert] = (temp_key, temp_cert)
        return temp_key.name, temp_cert.name
    args = parser.parse_args()
    pidfile = args.pidfile
    if pidfile:
        setRunning(pidfile)
    try:
        log = args.log
        formatter = logging.Formatter('%(asctime)s %(message)s')
        logger = logging.getLogger()
        if args.verbose:
            log_level = logging.DEBUG
        else:
            log_level = logging.INFO
        logger.setLevel(log_level)
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        if log:
            handler = logging.FileHandler(log)
            handler.setFormatter(formatter)
            logger.addHandler(handler)
            log_file = open(log)
            log_file.seek(0, 2)
        section_dict = {}
        configuration = ConfigParser.SafeConfigParser()
        configuration.readfp(args.configuration_file)
        for section in configuration.sections():
            if section == 'agent':
                continue
            section_dict[section] = section_entry_dict = dict(
                configuration.items(section))
            for key in ('request_kw', 'max_install_duration',
                        'max_destroy_duration', 'max_request_duration',
                        'max_uninstall_duration', 'computer_list'
                    ):
                if key in section_entry_dict:
                    try:
                        if isinstance(section_entry_dict[key], str) or \
                              isinstance(section_entry_dict[key], unicode):
                            section_entry_dict[key] = json.loads(
                                                  section_entry_dict[key])
                    except Exception as exc:
                        logger.error("Fail to load %s on %s" % (key, section_entry_dict))
                        raise
            if 'key' in section_entry_dict:
                key_file, cert_file = asFilenamePair(section_entry_dict['key'],
                    section_entry_dict['cert'])
                section_entry_dict['key'] = key_file
                section_entry_dict['cert'] = cert_file
            if "computer_list" in section_entry_dict:
                section_entry_dict["target_computer"] = \
                          random.choice(section_entry_dict["computer_list"])
        agent_parameter_dict = dict(configuration.items('agent'))
        # XXX: should node title be auto-generated by installation recipe ?
        # For example, using computer guid.
        node_title = agent_parameter_dict['node_title']
        test_title = agent_parameter_dict['test_title']
        project_title = agent_parameter_dict['project_title']
        task_distribution_tool = TaskDistributionTool(agent_parameter_dict[
            'report_url'])
        master_slap_connection_dict = {}
        test_result = task_distribution_tool.createTestResult(
            revision='',
            test_name_list=section_dict.keys(),
            node_title=node_title,
            allow_restart=True,
            test_title=test_title,
            project_title=project_title,
        )
        test_result.watcher_period = 300
        if log:
            test_result.addWatch(log, log_file, max_history_bytes=10000)
        assert test_result is not None
        test_mapping = TestMap(section_dict)
        logger.info("Running %s tests in parallel." % \
                      len(test_mapping.getComputerList()))

        ran_test_set = set()
        running_test_dict = {}
        more_tests = True
        logger.info('Starting Test Agent run %s ' % node_title)
        while True:
            # Get up to parallel_task_count tasks to execute
            while len(running_test_dict) < len(test_mapping.getComputerList())\
                    and more_tests:
                test_mapping.cleanUp()
                target_computer = test_mapping.getNextComputer([computer \
                        for _, _, computer in running_test_dict.itervalues()])

                test_line = test_result.start(
                    exclude_list= list(ran_test_set) + \
                           list(test_mapping.getExcludeList(target_computer)))

                logger.info("Test Line: %s " % test_line)
                logger.info("Ran Test Set: %s " % ran_test_set)
                logger.info("Running test dict: %s " % running_test_dict)
                logger.info("Target Computer: %s " % target_computer)
                if test_line is None:
                    test_mapping.dropComputer(target_computer)
                    if len(test_mapping.getComputerList()) == 0:
                        more_tests = False
                    continue
                test_name = test_line.name
                try:
                    section_entry_dict = section_dict[test_name]
                except KeyError:
                    # We don't know how to execute this test. Assume it doesn't
                    # exist anymore, and fail it in result.
                    test_line.stop(stderr='This test does not exist on test '
                        'node %s' % (node_title, ))
                    continue
                master_url = section_entry_dict['master_url']
                master_slap_connection_key = (master_url,
                    section_entry_dict.get('key'))
                try:
                    supply, order, rpc = master_slap_connection_dict[
                        master_slap_connection_key]
                except KeyError:
                    key = section_entry_dict.get('key')
                    cert = section_entry_dict.get('cert')
                    slap = slapos.slap.slap()
                    slap.initializeConnection(master_url, key, cert)
                    supply = slap.registerSupply()
                    order = slap.registerOpenOrder()
                    assert master_url.startswith('https:')
                    rpc = xmlrpclib.ServerProxy(master_url, allow_none=True,
                        transport=x509Transport(
                            {'key_file': key, 'cert_file': cert}))
                    master_slap_connection_dict[
                        master_slap_connection_key] = (supply, order, rpc)
                tester = SoftwareReleaseTester(
                    test_name + '_' + node_title + time.strftime(
                        '_%Y/%m/%d_%H:%M:%S_+0000', time.gmtime()),
                    logger,
                    rpc,
                    supply,
                    order,
                    section_entry_dict['url'],
                    section_entry_dict['target_computer'],
                    section_entry_dict['max_install_duration'],
                    section_entry_dict['max_uninstall_duration'],
                    section_entry_dict.get('request_kw'),
                    section_entry_dict.get('max_request_duration'),
                    section_entry_dict.get('max_destroy_duration'),
                )
                ran_test_set.add(test_name)
                running_test_dict[test_name] = (test_line, tester, target_computer)
            if not running_test_dict:
               break

            now = time.time()
            # Synchronise refreshes on watcher period, so it doesn't report a
            # stalled test node where we are actually still sleeping.
            # Change test_result.watcher_period outside this loop if you wish
            # to change sleep duration.
            next_deadline = now + test_result.watcher_period
            for section, (test_line, tester, target_computer) in running_test_dict.items():
                logger.info('Checking %s: %r...', section, tester)
                try:
                    deadline = tester.tic(now)
                except Exception:
                    logger.exception('Test execution fail for  %s' % (section))
                    test_line.stop(
                        test_count=1,
                        error_count=1,
                        failure_count=0,
                        skip_count=0,
                        stderr=traceback.format_exc(),
                    )
                    del running_test_dict[section]
                    try:
                        tester.teardown()
                    except slapos.slap.NotFoundError:
                        # This exception is ignored because we cannot
                        # Teardown if SR URL do not exist.
                        logger.exception('Fail and not found')
                        pass
                    except Exception:
                        logger.exception('teardown failed, human '
                            'assistance needed for cleanup')
                        raise
                else:
                    logger.info('%r', tester)
                    if deadline is None:
                        # TODO: report how long each step took.
                        logger.info('Test execution finished for  %s' % (section))
                        test_line.stop(
                            test_count=1,
                            error_count=0,
                            failure_count=0,
                            skip_count=0,
                        )
                        del running_test_dict[section]
                        try:
                            tester.teardown()
                        except slapos.slap.NotFoundError:
                            # This exception is ignored because we cannot
                            # Teardown if SR URL do not exist.
                            logger.exception('Fail and not found')
                            pass
                        except Exception:
                            logger.exception('teardown failed, human '
                                 'assistance needed for cleanup')
                            raise

                    else:
                        next_deadline = min(deadline, next_deadline)
            if running_test_dict:
                to_sleep = next_deadline - time.time()
                if to_sleep > 0:
                    logger.info('Sleeping %is...', to_sleep)
                    time.sleep(to_sleep)
                if not test_result.isAlive():
                    for _, tester, computer_id in running_test_dict.itervalues():
                        tester.teardown()
    finally:
        if pidfile:
            setFinished(pidfile)
        # Help interpreter get rid of AutoSTemp instances.
        key_file_dict.clear()

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

import os
import urllib
from flask import jsonify
from werkzeug import secure_filename
import shutil
import datetime
import hashlib
from utils import realpath, tail, isText
import re
import zipfile

class fileBrowser(object):
  """This class contain all bases function for file browser"""

  def __init__(self, config):
    self.config = config

  def listDirs(self, dir, all=False):
    """List elements of directory 'dir' taken"""
    html = ''
    html += 'var gsdirs = new Array();'
    html += 'var gsfiles = new Array();'
    dir = urllib.unquote(str(dir))
    realdir = realpath(self.config, dir)
    if not realdir:
      raise NameError('Could not load directory %s: Permission denied' % dir)

    ldir = sorted(os.listdir(realdir), key=str.lower)
    for f in ldir:
      if f.startswith('.') and not all: #do not display this file/folder
        continue
      ff = os.path.join(dir,f)
      realfile = os.path.join(realdir, f)
      mdate = datetime.datetime.fromtimestamp(os.path.getmtime(realfile)
                    ).strftime("%Y-%d-%m %I:%M")
      md5 = hashlib.md5(realfile).hexdigest()
      if not os.path.isdir(realfile):
        size = os.path.getsize(realfile)
      	regex = re.compile("(^.*)\.(.*)", re.VERBOSE)
      	ext = regex.sub(r'\2', f)
      	if ext == f:
      	  ext = "unknow"
      	else:
	      ext = str.lower(ext)
      	html += 'gsfiles.push(new gsItem("1", "' + f + '", "' + \
                  ff + '", "' + str(size) + '", "' + md5 + \
                  '", "' + ext + '", "' + mdate + '"));';
      else:
        html += 'gsdirs.push(new gsItem("2", "' + f + '", "' + \
                  ff + '", "0", "' + md5 + '", "dir", "' + mdate + '"));';
    return html


  def makeDirectory(self, dir, filename):
    """Create a directory"""
    realdir = realpath(self.config, urllib.unquote(dir))
    if not realdir:
      raise NameError('Could not load directory %s: Permission denied' % dir)
    folder = os.path.join(realdir, filename)
    if not os.path.exists(folder):
      os.mkdir(folder, 0744)
      return '{result: \'1\'}'
    else:
      return '{result: \'0\'}'

  def makeFile(self, dir, filename):
    """Create a file in a directory dir taken"""
    realdir = realpath(self.config, urllib.unquote(dir))
    if not realdir:
      raise NameError('Could not load directory %s: Permission denied' % dir)
    file = os.path.join(realdir, filename)
    if not os.path.exists(file):
      open(file, 'w').write('')
      return 'var responce = {result: \'1\'}'
    else:
      return '{result: \'0\'}'

  def deleteItem(self, dir, files):
    """Delete a list of files or directories"""
    realdir = realpath(self.config, urllib.unquote(dir))
    if not realdir:
      raise NameError('Could not load directory %s: Permission denied' % dir)
    lfiles = urllib.unquote(files).split(',,,')
    try:
      for file in lfiles:
        file = os.path.join(realdir, file)
        if not os.path.exists(file):
          continue #silent skip file....
        details = file.split('/')
        last = details[len(details) - 1]
        if last and last.startswith('.'):
          continue #cannot delete this file/directory, to prevent security
        if os.path.isdir(file):
          shutil.rmtree(file)
        else:
          os.unlink(file)
    except Exception, e:
      return str(e)
    return '{result: \'1\'}'

  def copyItem(self, dir, files, del_source=False):
    """Copy a list of files or directory to dir"""
    realdir = realpath(self.config, urllib.unquote(dir))
    if not realdir:
      raise NameError('Could not load directory %s: Permission denied' % dir)
    lfiles = urllib.unquote(files).split(',,,')
    try:
      for file in lfiles:
        realfile = realpath(self.config, file)
        if not realfile:
          raise NameError('Could not load file or directory %s: Permission denied' % file)
        #prepare destination file
        details = realfile.split('/')
        dest = os.path.join(realdir, details[len(details) - 1])
        if os.path.exists(dest):
          raise NameError('NOT ALLOWED OPERATION : File or directory already exist')
        if os.path.isdir(realfile):
          shutil.copytree(realfile, dest)
          if del_source:
            shutil.rmtree(realfile)
        else:
          shutil.copy(realfile, dest)
          if del_source:
            os.unlink(realfile)
    except Exception, e:
      return str(e)
    return '{result: \'1\'}'

  def rename(self, dir, filename, newfilename):
    """Rename file or directory to dir/filename"""
    realdir = realpath(self.config, urllib.unquote(dir))
    if not realdir:
      raise NameError('Could not load directory %s: Permission denied' % dir)
    realfile = realpath(self.config, urllib.unquote(filename))
    if not realfile:
      raise NameError('Could not load directory %s: Permission denied' % filename)
    tofile = os.path.join(realdir, newfilename)
    if not os.path.exists(tofile):
      os.rename(realfile, tofile)
      return '{result: \'1\'}'
    raise NameError('NOT ALLOWED OPERATION : File or directory already exist')

  def copyAsFile(self, dir, filename, newfilename):
    """Copy file or directory to dir/filename"""
    realdir = realpath(self.config, urllib.unquote(dir))
    if not realdir:
      raise NameError('Could not load directory %s: Permission denied' % dir)
    fromfile = os.path.join(realdir, filename)
    tofile = os.path.join(realdir, newfilename)
    if not os.path.exists(fromfile):
      raise NameError('NOT ALLOWED OPERATION : File or directory not exist')
    if not os.path.exists(tofile):
      shutil.copy(fromfile, tofile)
      return '{result: \'1\'}'
    raise NameError('NOT ALLOWED OPERATION : File or directory already exist')

  def uploadFile(self, dir, files):
    """Upload a list of file in directory dir"""
    realdir = realpath(self.config, urllib.unquote(dir))
    if not realdir:
      raise NameError('Could not load directory %s: Permission denied' % dir)
    for file in files:
      if files[file]:
        filename = secure_filename(files[file].filename)
        if not os.path.exists(os.path.join(dir, filename)):
          files[file].save(os.path.join(realdir, filename))
    return '{result: \'1\'}'

  def downloadFile(self, dir, filename):
    """Download file dir/filename"""
    realdir = realpath(self.config, urllib.unquote(dir))
    if not realdir:
      raise NameError('Could not load directory %s: Permission denied' % dir)
    file = os.path.join(realdir, urllib.unquote(filename))
    if not os.path.exists(file):
      raise NameError('NOT ALLOWED OPERATION : File or directory does not exist %s'
                      % os.path.join(dir, filename))
    return file

  def zipFile(self, dir, filename, newfilename):
    """Add filename to archive as newfilename"""
    realdir = realpath(self.config, urllib.unquote(dir))
    if not realdir:
      raise NameError('Could not load directory %s: Permission denied' % dir)
    tozip = os.path.join(realdir, newfilename)
    fromzip = os.path.join(realdir, filename)
    if not os.path.exists(fromzip):
      raise NameError('NOT ALLOWED OPERATION : File or directory not exist')
    if not os.path.exists(tozip):
      zip = zipfile.ZipFile(tozip, 'w', zipfile.ZIP_DEFLATED)
      if os.path.isdir(fromzip):
        rootlen = len(fromzip) + 1
        for base, dirs, files in os.walk(fromzip):
          for file in files:
            fn = os.path.join(base, file).encode("utf-8")
            zip.write(fn, fn[rootlen:])
      else:
        zip.write(fromzip)
      zip.close()
      return '{result: \'1\'}'
    raise NameError('NOT ALLOWED OPERATION : File or directory already exist')

  def unzipFile(self, dir, filename, newfilename):
    """Extract a zipped archive"""
    realdir = realpath(self.config, urllib.unquote(dir))
    if not realdir:
      raise NameError('Could not load directory %s: Permission denied' % dir)
    target = os.path.join(realdir, newfilename)
    archive = os.path.join(realdir, filename)
    if not os.path.exists(archive):
      raise NameError('NOT ALLOWED OPERATION : File or directory not exist')
    if not os.path.exists(target):
      zip = zipfile.ZipFile(archive)
      #member = zip.namelist()
      zip.extractall(target)
      #if len(member) > 1:
      #  zip.extractall(target)
      #else:
      #  zip.extract(member[0], newfilename)
      return '{result: \'1\'}'
    raise NameError('NOT ALLOWED OPERATION : File or directory already exist')

  def readFile(self, dir, filename, truncate=0):
    """Read file dir/filename and return content"""
    realfile = realpath(self.config, os.path.join(urllib.unquote(dir),
                        urllib.unquote(filename)))
    if not realfile:
      raise NameError('Could not load directory %s: Permission denied' % dir)
    if not isText(realfile):
      return "FILE ERROR: Cannot display binary file, please open a text file only!"
    if truncate == 0:
      return open(realfile, 'r').read()
    else:
      tail(open(realfile, 'r'), 0)
